from functools import partial

import numpy as np
import pytest

from guacamol.score_modifier import LinearModifier, SquaredModifier, AbsoluteScoreModifier, GaussianModifier, \
    MinGaussianModifier, MaxGaussianModifier, ThresholdedLinearModifier, ClippedScoreModifier, \
    SmoothClippedScoreModifier, ChainedModifier

scalar_value = 8.343
value_array = np.array([[-3.3, 0, 5.5],
                        [0.011, 2.0, -33]])


def test_linear_function_default():
    f = LinearModifier()

    assert f(scalar_value) == scalar_value
    assert np.array_equal(f(value_array), value_array)


def test_linear_function_with_slope():
    slope = 3.3
    f = LinearModifier(slope=slope)

    assert f(scalar_value) == slope * scalar_value
    assert np.array_equal(f(value_array), slope * value_array)


def test_squared_function():
    target_value = 5.555
    coefficient = 0.123
    f = SquaredModifier(target_value=target_value, coefficient=coefficient)

    expected_scalar = 1.0 - coefficient * (target_value - scalar_value) ** 2
    expected_array = 1.0 - coefficient * np.square(target_value - value_array)

    assert f(scalar_value) == expected_scalar
    assert np.array_equal(f(value_array), expected_array)


def test_absolute_function():
    target_value = 5.555
    f = AbsoluteScoreModifier(target_value=target_value)

    expected_scalar = 1.0 - abs(target_value - scalar_value)
    expected_array = 1.0 - np.abs(target_value - value_array)

    assert f(scalar_value) == expected_scalar
    assert np.array_equal(f(value_array), expected_array)


def gaussian(x, mu, sig):
    return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))


def test_gaussian_function():
    mu = -1.223
    sigma = 0.334

    f = GaussianModifier(mu=mu, sigma=sigma)

    assert f(mu) == 1.0
    assert f(scalar_value) == gaussian(scalar_value, mu, sigma)
    assert np.allclose(f(value_array), gaussian(value_array, mu, sigma))


def test_min_gaussian_function():
    mu = -1.223
    sigma = 0.334

    f = MinGaussianModifier(mu=mu, sigma=sigma)

    assert f(mu) == 1.0

    low_value = -np.inf
    large_value = np.inf

    assert f(low_value) == 1.0
    assert f(large_value) == 0.0

    full_gaussian = partial(gaussian, mu=mu, sig=sigma)
    min_gaussian_lambda = lambda x: 1.0 if x < mu else full_gaussian(x)
    min_gaussian = np.vectorize(min_gaussian_lambda)

    assert f(scalar_value) == min_gaussian(scalar_value)
    assert np.allclose(f(value_array), min_gaussian(value_array))


def test_max_gaussian_function():
    mu = -1.223
    sigma = 0.334

    f = MaxGaussianModifier(mu=mu, sigma=sigma)

    assert f(mu) == 1.0

    low_value = -np.inf
    large_value = np.inf

    assert f(low_value) == 0.0
    assert f(large_value) == 1.0

    full_gaussian = partial(gaussian, mu=mu, sig=sigma)
    max_gaussian_lambda = lambda x: 1.0 if x > mu else full_gaussian(x)
    max_gaussian = np.vectorize(max_gaussian_lambda)

    assert f(scalar_value) == max_gaussian(scalar_value)
    assert np.allclose(f(value_array), max_gaussian(value_array))


def test_tanimoto_threshold_function():
    threshold = 5.555
    f = ThresholdedLinearModifier(threshold=threshold)

    large_value = np.inf

    assert f(large_value) == 1.0
    assert f(threshold) == 1.0

    expected_array = np.minimum(value_array, threshold) / threshold
    assert np.array_equal(f(value_array), expected_array)


def test_clipped_function():
    min_x = 4.4
    max_x = 8.8
    min_score = -3.3
    max_score = 9.2

    modifier = ClippedScoreModifier(upper_x=max_x, lower_x=min_x, high_score=max_score, low_score=min_score)

    # values smaller than min_x should be assigned min_score
    for x in [-2, 0, 4, 4.4]:
        assert modifier(x) == min_score

    # values larger than max_x should be assigned min_score
    for x in [8.8, 9.0, 1000]:
        assert modifier(x) == max_score

    # values in between are interpolated
    slope = (max_score - min_score) / (max_x - min_x)
    for x in [4.4, 4.8, 5.353, 8.034, 8.8]:
        dx = x - min_x
        dy = dx * slope
        assert modifier(x) == pytest.approx(min_score + dy)


def test_clipped_function_inverted():
    # The clipped function also works for decreasing scores
    max_x = 4.4
    min_x = 8.8
    min_score = -3.3
    max_score = 9.2

    modifier = ClippedScoreModifier(upper_x=max_x, lower_x=min_x, high_score=max_score, low_score=min_score)

    # values smaller than max_x should be assigned the maximal score
    for x in [-2, 0, 4, 4.4]:
        assert modifier(x) == max_score

    # values larger than min_x should be assigned min_score
    for x in [8.8, 9.0, 1000]:
        assert modifier(x) == min_score

    # values in between are interpolated
    slope = (max_score - min_score) / (max_x - min_x)
    for x in [4.4, 4.8, 5.353, 8.034, 8.8]:
        dx = x - min_x
        dy = dx * slope
        assert modifier(x) == pytest.approx(min_score + dy)


def test_thresholded_is_special_case_of_clipped_for_positive_input():
    threshold = 4.584
    thresholded_modifier = ThresholdedLinearModifier(threshold=threshold)
    clipped_modifier = ClippedScoreModifier(upper_x=threshold)

    values = np.array([0, 2.3, 8.545, 3.23, 0.12, 55.555])

    assert np.allclose(thresholded_modifier(values), clipped_modifier(values))


def test_smooth_clipped():
    min_x = 4.4
    max_x = 8.8
    min_score = -3.3
    max_score = 9.2

    modifier = SmoothClippedScoreModifier(upper_x=max_x, lower_x=min_x, high_score=max_score, low_score=min_score)

    # assert that the slope in the middle is correct

    middle_x = (min_x + max_x) / 2
    delta = 1e-5
    vp = modifier(middle_x + delta)
    vm = modifier(middle_x - delta)

    slope = (vp - vm) / (2 * delta)
    expected_slope = (max_score - min_score) / (max_x - min_x)

    assert slope == pytest.approx(expected_slope)

    # assert behavior at +- infinity

    assert modifier(1e5) == pytest.approx(max_score)
    assert modifier(-1e5) == pytest.approx(min_score)


def test_smooth_clipped_inverted():
    # The smooth clipped function also works for decreasing scores
    max_x = 4.4
    min_x = 8.8
    min_score = -3.3
    max_score = 9.2

    modifier = SmoothClippedScoreModifier(upper_x=max_x, lower_x=min_x, high_score=max_score, low_score=min_score)

    # assert that the slope in the middle is correct

    middle_x = (min_x + max_x) / 2
    delta = 1e-5
    vp = modifier(middle_x + delta)
    vm = modifier(middle_x - delta)

    slope = (vp - vm) / (2 * delta)
    expected_slope = (max_score - min_score) / (max_x - min_x)

    assert slope == pytest.approx(expected_slope)

    # assert behavior at +- infinity

    assert modifier(1e5) == pytest.approx(min_score)
    assert modifier(-1e5) == pytest.approx(max_score)


def test_chained_modifier():
    linear = LinearModifier(slope=2)
    squared = SquaredModifier(10.0)

    chained_1 = ChainedModifier([linear, squared])
    chained_2 = ChainedModifier([squared, linear])

    expected_1 = 1.0 - np.square(10.0 - (2 * scalar_value))
    expected_2 = 2 * (1.0 - np.square(10.0 - scalar_value))

    assert chained_1(scalar_value) == expected_1
    assert chained_2(scalar_value) == expected_2



from math import sqrt
from typing import List

import pytest

from guacamol.common_scoring_functions import IsomerScoringFunction, SMARTSScoringFunction
from guacamol.score_modifier import GaussianModifier
from guacamol.scoring_function import BatchScoringFunction, ArithmeticMeanScoringFunction, GeometricMeanScoringFunction
from guacamol.utils.math import geometric_mean


class MockScoringFunction(BatchScoringFunction):
    """
    Mock scoring function that returns values from an array given in the constructor.
    """

    def __init__(self, values: List[float]) -> None:
        super().__init__()
        self.values = values
        self.index = 0

    def raw_score_list(self, smiles_list: List[str]) -> List[float]:
        start = self.index
        self.index += len(smiles_list)
        end = self.index
        return self.values[start:end]


def test_isomer_scoring_function_uses_geometric_mean_by_default():
    scoring_function = IsomerScoringFunction('C2H4')
    assert scoring_function.mean_function == geometric_mean


def test_isomer_scoring_function_returns_one_for_correct_molecule():
    c11h24_arithmetic = IsomerScoringFunction('C11H24', mean_function='arithmetic')
    c11h24_geometric = IsomerScoringFunction('C11H24', mean_function='geometric')

    # all those smiles fit the formula C11H24
    smiles1 = 'CCCCCCCCCCC'
    smiles2 = 'CC(CCC)CCCCCC'
    smiles3 = 'CCCCC(CC(C)CC)C'

    assert c11h24_arithmetic.score(smiles1) == 1.0
    assert c11h24_arithmetic.score(smiles2) == 1.0
    assert c11h24_arithmetic.score(smiles3) == 1.0
    assert c11h24_geometric.score(smiles1) == 1.0
    assert c11h24_geometric.score(smiles2) == 1.0
    assert c11h24_geometric.score(smiles3) == 1.0


def test_isomer_scoring_function_penalizes_additional_atoms():
    c11h24_arithmetic = IsomerScoringFunction('C11H24', mean_function='arithmetic')
    c11h24_geometric = IsomerScoringFunction('C11H24', mean_function='geometric')

    # all those smiles are C11H24O
    smiles1 = 'CCCCCCCCCCCO'
    smiles2 = 'CC(CCC)COCCCCC'
    smiles3 = 'CCCCOC(CC(C)CC)C'

    # the penalty corresponds to a deviation of 1.0 from the gaussian modifier for the total number of atoms
    n_atoms_score = GaussianModifier(mu=0, sigma=2)(1.0)
    c_score = 1.0
    h_score = 1.0
    expected_score_arithmetic = (n_atoms_score + c_score + h_score) / 3.0
    expected_score_geometric = (n_atoms_score * c_score * h_score) ** (1 / 3)

    assert c11h24_arithmetic.score(smiles1) == pytest.approx(expected_score_arithmetic)
    assert c11h24_arithmetic.score(smiles2) == pytest.approx(expected_score_arithmetic)
    assert c11h24_arithmetic.score(smiles3) == pytest.approx(expected_score_arithmetic)
    assert c11h24_geometric.score(smiles1) == pytest.approx(expected_score_geometric)
    assert c11h24_geometric.score(smiles2) == pytest.approx(expected_score_geometric)
    assert c11h24_geometric.score(smiles3) == pytest.approx(expected_score_geometric)


def test_isomer_scoring_function_penalizes_incorrect_number_atoms():
    c11h24_arithmetic = IsomerScoringFunction('C12H24', mean_function='arithmetic')
    c11h24_geometric = IsomerScoringFunction('C12H24', mean_function='geometric')

    # all those smiles fit the formula C11H24O
    smiles1 = 'CCCCCCCCOCCC'
    smiles2 = 'CC(CCOC)CCCCCC'
    smiles3 = 'COCCCC(CC(C)CC)C'

    # the penalty corresponds to a deviation of 1.0 from the gaussian modifier in the number of C atoms
    c_score = GaussianModifier(mu=0, sigma=1)(1.0)
    n_atoms_score = 1.0
    h_score = 1.0
    expected_score_arithmetic = (n_atoms_score + c_score + h_score) / 3.0
    expected_score_geometric = (n_atoms_score * c_score * h_score) ** (1 / 3)

    assert c11h24_arithmetic.score(smiles1) == pytest.approx(expected_score_arithmetic)
    assert c11h24_arithmetic.score(smiles2) == pytest.approx(expected_score_arithmetic)
    assert c11h24_arithmetic.score(smiles3) == pytest.approx(expected_score_arithmetic)
    assert c11h24_geometric.score(smiles1) == pytest.approx(expected_score_geometric)
    assert c11h24_geometric.score(smiles2) == pytest.approx(expected_score_geometric)
    assert c11h24_geometric.score(smiles3) == pytest.approx(expected_score_geometric)


def test_isomer_scoring_function_invalid_molecule():
    sf = IsomerScoringFunction('C60')

    assert sf.score('CCCinvalid') == sf.corrupt_score


def test_smarts_function():
    mol1 = 'COc1cc(N(C)CCN(C)C)c(NC(=O)C=C)cc1Nc2nccc(n2)c3cn(C)c4ccccc34'
    mol2 = 'Cc1c(C)c2OC(C)(COc3ccc(CC4SC(=O)NC4=O)cc3)CCc2c(C)c1O'
    smarts = '[#7;h1]c1ncccn1'

    scofu1 = SMARTSScoringFunction(target=smarts)
    scofu_inv = SMARTSScoringFunction(target=smarts, inverse=True)

    assert scofu1.score(mol1) == 1.0
    assert scofu1.score(mol2) == 0.0
    assert scofu_inv.score(mol1) == 0.0
    assert scofu_inv.score(mol2) == 1.0

    assert scofu1.score_list([mol1])[0] == 1.0
    assert scofu1.score_list([mol2])[0] == 0.0


def test_arithmetic_mean_scoring_function():
    # define a scoring function returning the mean from two mock functions
    # and assert that it returns the correct values.

    weight_1 = 0.4
    weight_2 = 0.6

    mock_values_1 = [0.232, 0.665, 0.0, 1.0, 0.993]
    mock_values_2 = [0.010, 0.335, 0.8, 0.3, 0.847]

    mock_1 = MockScoringFunction(mock_values_1)
    mock_2 = MockScoringFunction(mock_values_2)

    scoring_function = ArithmeticMeanScoringFunction(scoring_functions=[mock_1, mock_2],
                                                     weights=[weight_1, weight_2])

    smiles = ['CC'] * 5

    scores = scoring_function.score_list(smiles)
    expected = [weight_1 * v1 + weight_2 * v2 for v1, v2 in zip(mock_values_1, mock_values_2)]

    assert scores == expected


def test_geometric_mean_scoring_function():
    # define a scoring function returning the geometric mean from two mock functions
    # and assert that it returns the correct values.

    mock_values_1 = [0.232, 0.665, 0.0, 1.0, 0.993]
    mock_values_2 = [0.010, 0.335, 0.8, 0.3, 0.847]

    mock_1 = MockScoringFunction(mock_values_1)
    mock_2 = MockScoringFunction(mock_values_2)

    scoring_function = GeometricMeanScoringFunction(scoring_functions=[mock_1, mock_2])

    smiles = ['CC'] * 5

    scores = scoring_function.score_list(smiles)
    expected = [sqrt(v1 * v2) for v1, v2 in zip(mock_values_1, mock_values_2)]

    assert scores == expected



from guacamol.distribution_learning_benchmark import ValidityBenchmark, UniquenessBenchmark, NoveltyBenchmark, \
    KLDivBenchmark
from guacamol.assess_distribution_learning import _assess_distribution_learning
from .mock_generator import MockGenerator
import numpy as np
import tempfile
from os.path import join


def test_validity_does_not_penalize_duplicates():
    generator = MockGenerator(['CCC', 'CCC'])
    benchmark = ValidityBenchmark(number_samples=2)

    assert benchmark.assess_model(generator).score == 1.0


def test_validity_score_is_proportion_of_valid_molecules():
    generator = MockGenerator(['CCC', 'CC(CC)C', 'invalidMolecule'])
    benchmark = ValidityBenchmark(number_samples=3)

    assert benchmark.assess_model(generator).score == 2.0 / 3.0


def test_uniqueness_penalizes_duplicates():
    generator = MockGenerator(['CCC', 'CCC', 'CCC'])
    benchmark = UniquenessBenchmark(number_samples=3)

    assert benchmark.assess_model(generator).score == 1.0 / 3.0


def test_uniqueness_penalizes_duplicates_with_different_smiles_strings():
    generator = MockGenerator(['C(O)C', 'CCO', 'OCC'])
    benchmark = UniquenessBenchmark(number_samples=3)

    assert benchmark.assess_model(generator).score == 1.0 / 3.0


def test_uniqueness_does_not_penalize_invalid_molecules():
    generator = MockGenerator(['C(O)C', 'invalid1', 'invalid2', 'CCC', 'NCCN'])
    benchmark = UniquenessBenchmark(number_samples=3)

    assert benchmark.assess_model(generator).score == 1.0


def test_novelty_score_is_zero_if_no_molecule_is_new():
    molecules = ['CCOCC', 'NNNNONNN', 'C=CC=C']
    generator = MockGenerator(molecules)
    benchmark = NoveltyBenchmark(number_samples=3, training_set=molecules)

    assert benchmark.assess_model(generator).score == 0.0


def test_novelty_score_is_one_if_all_molecules_are_new():
    generator = MockGenerator(['CCOCC', 'NNNNONNN', 'C=CC=C'])
    benchmark = NoveltyBenchmark(number_samples=3, training_set=['CO', 'CC'])

    assert benchmark.assess_model(generator).score == 1.0


def test_novelty_score_does_not_penalize_duplicates():
    generator = MockGenerator(['CCOCC', 'O(CC)CC', 'C=CC=C', 'CC'])
    benchmark = NoveltyBenchmark(number_samples=3, training_set=['CO', 'CC'])

    # Gets 2 out of 3: one of the duplicated molecules is ignored, so the sampled molecules are
    # ['CCOCC', 'C=CC=C', 'CC'],  and 'CC' is not novel
    assert benchmark.assess_model(generator).score == 2.0 / 3.0


def test_novelty_score_penalizes_invalid_molecules():
    generator = MockGenerator(['CCOCC', 'invalid1', 'invalid2', 'CCCC', 'CC'])
    benchmark = NoveltyBenchmark(number_samples=3, training_set=['CO', 'CC'])

    assert benchmark.assess_model(generator).score == 2.0 / 3.0


def test_KLdiv_benchmark_same_dist():
    generator = MockGenerator(['CCOCC', 'NNNNONNN', 'C=CC=C'])
    benchmark = KLDivBenchmark(number_samples=3, training_set=['CCOCC', 'NNNNONNN', 'C=CC=C'])
    result = benchmark.assess_model(generator)
    print(result.metadata)
    assert np.isclose(result.score, 1.0, )


def test_KLdiv_benchmark_different_dist():
    generator = MockGenerator(['CCOCC', 'NNNNONNN', 'C=CC=C'])
    benchmark = KLDivBenchmark(number_samples=3, training_set=['FCCOCC', 'CC(CC)CCCCNONNN', 'C=CC=O'])
    result = benchmark.assess_model(generator)
    print(result.metadata)

    assert result.metadata['number_samples'] == 3
    assert result.metadata.get('kl_divs') is not None
    assert result.metadata['kl_divs'].get('BertzCT') > 0
    assert result.metadata['kl_divs'].get('MolLogP', None) > 0
    assert result.metadata['kl_divs'].get('MolWt', None) > 0
    assert result.metadata['kl_divs'].get('TPSA', None) > 0
    assert result.metadata['kl_divs'].get('NumHAcceptors', None) > 0
    assert result.metadata['kl_divs'].get('NumHDonors', None) > 0
    assert result.metadata['kl_divs'].get('NumRotatableBonds', None) > 0
    assert result.score < 1.0


def test_distribution_learning_suite_v1():
    generator = MockGenerator(
        ['CCl', 'CCOCCCl', 'ClCCF', 'CCCOCCOCCCO', 'CF', 'CCOCC', 'CCF', 'CCCOCC', 'NNNNONNN', 'C=CC=C'] * 10)

    mock_chembl = ['FCCOCC', 'C=CC=O', 'CCl', 'CCOCCCl', 'ClCCF', 'CCCOCCOCCCO', 'CF', 'CCOCC',
                   'CCF']

    temp_dir = tempfile.mkdtemp()
    smiles_path = join(temp_dir, 'mock.smiles')
    with open(smiles_path, 'w') as f:
        for i in mock_chembl:
            f.write(f'{i}\n')
        f.close()

    json_path = join(temp_dir, 'output.json')

    _assess_distribution_learning(model=generator,
                                  chembl_training_file=smiles_path,
                                  json_output_file=json_path,
                                  benchmark_version='v1',
                                  number_samples=4)

    with open(json_path, 'r') as f:
        print(f.read())



from guacamol.utils.sampling_helpers import sample_valid_molecules, sample_unique_molecules
from .mock_generator import MockGenerator


def test_sample_valid_molecules_for_valid_only():
    generator = MockGenerator(['CCCC', 'CC'])

    mols = sample_valid_molecules(generator, 2)

    assert mols == ['CCCC', 'CC']


def test_sample_valid_molecules_with_invalid_molecules():
    generator = MockGenerator(['invalid', 'invalid', 'invalid', 'CCCC', 'invalid', 'CC'])

    mols = sample_valid_molecules(generator, 2)

    assert mols == ['CCCC', 'CC']


def test_sample_valid_molecules_if_not_enough_valid_generated():
    # does not raise an exception if
    molecules = ['invalid' for _ in range(20)]
    molecules[-1] = 'CC'
    molecules[-2] = 'CN'
    generator = MockGenerator(molecules)

    # samples a max of 9*2 molecules and just does not sample the good ones
    # in this case the list of generated molecules is empty
    assert not sample_valid_molecules(generator, 2, max_tries=9)

    # with a max of 10*2 molecules two valid molecules can be sampled
    generator = MockGenerator(molecules)
    mols = sample_valid_molecules(generator, 2)
    assert mols == ['CN', 'CC']


def test_sample_unique_molecules_for_valid_only():
    generator = MockGenerator(['CCCC', 'CC'])

    mols = sample_unique_molecules(generator, 2)

    assert mols == ['CCCC', 'CC']


def test_sample_unique_molecules_with_invalid_molecules():
    generator = MockGenerator(['invalid1', 'invalid2', 'inv3', 'CCCC', 'CC'])

    mols = sample_unique_molecules(generator, 2)

    assert mols == ['CCCC', 'CC']


def test_sample_unique_molecules_with_duplicate_molecules():
    generator = MockGenerator(['CO', 'C(O)', 'CCCC', 'CC'])

    mols = sample_unique_molecules(generator, 2)

    assert mols == ['CO', 'CCCC']


def test_sample_unique_molecules_if_not_enough_unique_generated():
    # does not raise an exception if
    molecules = ['CO' for _ in range(20)]
    molecules[-1] = 'CC'
    generator = MockGenerator(molecules)

    # samples a max of 9*2 molecules and just does not sample the other molecule
    # in this case the list of generated molecules contains just 'CO'
    mols = sample_unique_molecules(generator, 2, max_tries=9)
    assert mols == ['CO']

    # with a max of 10*2 molecules two valid molecules can be sampled
    generator = MockGenerator(molecules)
    mols = sample_unique_molecules(generator, 2)
    assert mols == ['CO', 'CC']



from typing import Optional, List

import pytest
from rdkit import Chem

from guacamol.goal_directed_benchmark import GoalDirectedBenchmark
from guacamol.goal_directed_generator import GoalDirectedGenerator
from guacamol.goal_directed_score_contributions import uniform_specification
from guacamol.scoring_function import ScoringFunctionBasedOnRdkitMol, ScoringFunction


class MockScoringFunction(ScoringFunctionBasedOnRdkitMol):
    """
    For testing purposes: scoring function that returns 0.1 * (number of atoms)
    """

    def score_mol(self, mol: Chem.Mol) -> float:
        return 0.1 * mol.GetNumAtoms()


class MockGenerator(GoalDirectedGenerator):
    """
    Mock generator that returns pre-defined molecules
    """
    def __init__(self, molecules: List[str]) -> None:
        self.molecules = molecules

    def generate_optimized_molecules(self, scoring_function: ScoringFunction, number_molecules: int,
                                     starting_population: Optional[List[str]] = None) -> List[str]:
        assert number_molecules == len(self.molecules)
        return self.molecules


def test_removes_duplicates():
    """
    Assert that duplicated molecules (even with different SMILES strings) are considered only once.
    """
    top3 = uniform_specification(3)
    benchmark = GoalDirectedBenchmark('benchmark', MockScoringFunction(), top3)
    generator = MockGenerator(['OCC', 'CCO', 'C(O)C'])

    individual_mock_score = 0.3

    assert benchmark.assess_model(generator).score == pytest.approx(individual_mock_score / 3)


def test_removes_invalid_molecules():
    top3 = uniform_specification(3)
    benchmark = GoalDirectedBenchmark('benchmark', MockScoringFunction(), top3)
    generator = MockGenerator(['OCC', 'invalid', 'invalid2'])

    individual_mock_score = 0.3

    assert benchmark.assess_model(generator).score == pytest.approx(individual_mock_score / 3)


def test_correct_score_averaging():
    top3 = uniform_specification(3)
    benchmark = GoalDirectedBenchmark('benchmark', MockScoringFunction(), top3)
    generator = MockGenerator(['OCC', 'CCCCOCCCC', 'C'])

    expected_score = (0.3 + 0.9 + 0.1) / 3

    assert benchmark.assess_model(generator).score == pytest.approx(expected_score)


def test_correct_score_with_multiple_contributions():
    """
    Verify that 0.5 * (top1 + top3) delivers the correct result
    """
    specification = uniform_specification(1, 3)
    benchmark = GoalDirectedBenchmark('benchmark', MockScoringFunction(), specification)
    generator = MockGenerator(['OCC', 'CCCCOCCCC', 'C'])

    top3 = (0.3 + 0.9 + 0.1) / 3
    top1 = 0.9
    expected_score = (top1 + top3) / 2

    assert benchmark.assess_model(generator).score == pytest.approx(expected_score)
